/* -*- Mode: C; tab-width: 8; c-basic-offset: 8; indent-tabs-mode: t; -*- */

#ifndef DEBUG_H
#define DEBUG_H

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/**
 * Useful debug macros, define DEBUGRR to turn on debugging
 *
 * Caution: never define NDEBUG as it turns off assertions!
 *
 */

//#define DEBUGRR

#ifdef DEBUGRR
#define debug(M, ...) fprintf(stderr, "[DEBUG] (%s) " M "\n", __FUNCTION__, ##__VA_ARGS__);
#define do_debug(C) C
#else
#define debug(M, ...)
#define do_debug(C)
#endif

#define clean_errno() (errno == 0 ? "None" : strerror(errno))

#define fatal(M, ...) do { fprintf(stderr, "[FATAL] (%s:%d:%s: errno: %s) " M "\n", __FILE__, __LINE__, __FUNCTION__, clean_errno(), ##__VA_ARGS__); abort(); } while (0)

#define log_err(M, ...) fprintf(stderr, "[ERROR] (%s:%d:%s: errno: %s) " M "\n", __FILE__, __LINE__, __FUNCTION__, clean_errno(), ##__VA_ARGS__)

#define log_warn(M, ...) fprintf(stderr, "[WARN] (%s: errno: %s) " M "\n", __FUNCTION__, clean_errno(), ##__VA_ARGS__)

#define log_info(M, ...) fprintf(stderr, "[INFO] (%s) " M "\n", __FUNCTION__, ##__VA_ARGS__)

#define check(A, M, ...) if(!(A)) { log_err(M, ##__VA_ARGS__); errno=0; goto error; }

#define sentinel(M, ...){ log_err(M, ##__VA_ARGS__); errno=0; goto error; }

#define check_mem(A) check((A), "Out of memory.")

#define check_debug(A, M, ...) if(!(A)) { debug(M, ##__VA_ARGS__); errno=0; goto error; }

#endif
